import pytest

from njupt import Zhengfang


@pytest.fixture(scope='session')
def zhengfang():
    zhengfang = Zhengfang()
    zhengfang.account = 'account'
    zhengfang.verify = True
    return zhengfang



import json
from io import BytesIO

import requests
from PIL import Image

from njupt import Library
from njupt.utils.captchas import build_vector
from njupt.utils.captchas.library.libray_captcha import LibraryCaptcha

if __name__ == "__main__":
    image_data = []

    for i in range(10):
        r = requests.get(Library.URLs.CAPTCHA)
        im = Image.open(BytesIO(r.content))
        im.show()
        captcha = LibraryCaptcha(image=im)
        letters = input()
        for letter, im in zip(letters, captcha.handle_split_image()):
            vector = build_vector(im)
            image_data.append({letter: vector})

    with open('image_data.json', 'w') as f:
        json.dump(image_data, f)



from datetime import datetime

import responses

from tests.conftest import mock_response


@mock_response(file_name='htmls/zhengfang/list_optional_courses.html')
@responses.activate
def test_list_optional_courses(zhengfang):
    optional_courses = zhengfang.list_optional_courses()
    assert len(optional_courses) == 2
    assert optional_courses[0]['学分'] == 2
    assert optional_courses[1]['余量'] == 150


@mock_response(file_name='htmls/zhengfang/get_grades.html')
@responses.activate
def test_list_exam_grades(zhengfang):
    grades = zhengfang.list_exam_grades()
    assert len(grades) == 6
    assert grades[0]['考试日期'] == datetime(2015, 12, 19)
    assert grades[0]['学期'] == 1


@mock_response(method='POST', file_name='htmls/zhengfang/scores.html')
@mock_response(method='GET', file_name='htmls/zhengfang/scores.html')
@responses.activate
def test_get_gpa(zhengfang):
    gpa = zhengfang.get_gpa()
    assert gpa == 3.52


@mock_response(method='POST', file_name='htmls/zhengfang/scores.html')
@mock_response(method='GET', file_name='htmls/zhengfang/scores.html')
@responses.activate
def test_list_exam_scores(zhengfang):
    exam_scores = zhengfang.list_exam_scores()
    assert exam_scores[0]['学分'] == 2.0
    assert exam_scores[0]['成绩'] == 81
    assert not exam_scores[0]['绩点']
    assert exam_scores[8]['成绩'] == "优秀"



# encoding: utf-8
import base64
import datetime
import json
import re

from njupt.exceptions import NjuptException, AuthenticationException
from njupt.base import API
from njupt.utils import login_required
from njupt.utils.captchas.card import CardCaptcha

AIDS = {
    'elec_xianlin': '0030000000005101',
    'elec_sanpailou': '0030000000005102',
    'net': '0030000000000301'
}


class Card(API):
    """
    一卡通相关接口

    :param str account: 一卡通卡号
    :param str password: 一卡通查询密码

    :raise: :class:`njupt.exceptions.AuthenticationException`

    >>> card = Card(account='B15080500', password='sssss')
    """

    class URLs:
        HOST = "http://yktapp.njupt.edu.cn:8070"
        # 登录地址
        LOGIN = HOST + '/Login/LoginBySnoQuery'
        # 一卡通验证码
        CAPTCHA = HOST + '/Login/GetValidateCode'
        # 一卡通基础信息
        INFO = HOST + '/User/GetCardInfoByAccountNoParm'
        # 一卡通消费记录
        BILL = HOST + '/Report/GetMyBill'
        # 一卡通充值
        RECHARGE = HOST + '/User/Account_Pay'
        # 一卡通Dr.com信息
        COMMON = HOST + '/Tsm/TsmCommon'
        # 一卡通网费充值
        NET_PAY = HOST + '/Tsm/Net_Pay'
        # 电费充值
        ELEC_PAY = HOST + '/Tsm/Elec_Pay'

    def __init__(self, account=None, password=None):
        super(Card, self).__init__()
        self.aid = None
        # 该系统使用aid区分业务类型
        self.inner_account = None
        if account and password:
            self.account = account
            self._login(account, password)

    def _login(self, account, password):
        captcha_code = CardCaptcha(self.get_image(self.URLs.CAPTCHA)).crack()
        data = {
            "sno": account,
            "pwd": base64.b64encode(str(password).encode("utf8")),
            "ValiCode": captcha_code,
            "remember": 1,
            "uclass": 1,
            "zqcode": "",
            "json": True,
        }

        result = self._login_execute(url=self.URLs.LOGIN, data=data)
        if result['code'] == 2:
            # 如果验证码错误，尝试递归重复登录
            return self._login(account, password)
        result['success'] = not result['code']
        if result['success']:
            self.verify = True
        else:
            raise AuthenticationException(result['msg'])
        return result

    def _login_execute(self, url=None, data=None):
        result = self.get_json(url=url, data=data, method="post")
        if result['IsSucceed']:
            return {'code': 0, 'msg': '登录成功'}
        else:
            if result['Msg'] == "验证码错误":
                return {'code': 2, 'msg': '验证码错误'}
            else:
                return {'code': 1, 'msg': result['Msg']}

    def _get_inner_account(self):
        """
        获取对应的内部账号，部分接口参数需要

        :return: 一卡通号对应的系统内部账号
        """
        if not self.inner_account:
            info = self.get_json(self.URLs.INFO, method='post', data={'json': True})
            result = json.loads(info['Msg'])
            self.inner_account = result['query_card']['card'][0]['account']
        return self.inner_account

    @login_required
    def get_balance(self):
        """
        查询余额

        :return: 包含 过度余额、到账余额、总余额的dict
        :rtype: dict of (str, float)

        >>> card.get_balance()
        {'balance': 10.01, 'unsettle_balance': 0.01, 'total': 10.02}
        """
        info = self.get_json(self.URLs.INFO, method='post', data={'json': True})
        result = json.loads(info['Msg'])
        balance = int(result['query_card']['card'][0]['db_balance']) / 100  # 到账余额
        unsettle_balance = int(result['query_card']['card'][0]['unsettle_amount']) / 100  # 过渡余额
        return {
            'balance': balance,  # 到账余额
            'unsettle_balance': unsettle_balance,  # 过渡余额
            'total': balance + unsettle_balance,  # 总余额
        }

    @login_required
    def get_net_balance(self):
        """
        获取Dr.com的上网费用余额

        :return: dr.com的余额
        :rtype: float

        >>> card.get_net_balance()
        2.33
        """
        data = {
            'jsondata': '{"query_net_info": {"aid": "%s", "account": "%s", "payacc": ""}}'  # 无法使用format格式化
                        % (AIDS['net'], self._get_inner_account()),
            'funname': 'synjones.onecard.query.net.info',
            'json': True,
        }
        result = self.get_json(url=self.URLs.COMMON, data=data, method='post')
        result = json.loads(result['Msg'])
        return float(re.search(r'余额(\d*\.\d*)元', result['query_net_info']['errmsg']).groups()[0])

    @login_required
    def recharge(self, amount):
        """
        从绑定的银行卡中扣款充值余额

        :param amount: 充值金额, 单位为元
        :type amount: float or int
        :return: 充值结果
        :rtype: dict

        >>> card.recharge(10)
        {'success': True, 'code': 0, 'msg': '转账成功'}
        """
        data = {
            'account': self._get_inner_account(),
            'acctype': '###',
            'tranamt': int(100 * amount),
            'qpwd': '',
            'paymethod': '2',
            'paytype': '使用绑定的默认账号',
            'client_type': 'web',
            'json': True,
        }
        info = self.get_json(self.URLs.RECHARGE, 'post', data)
        info = json.loads(info['Msg'])['transfer']
        result = {
            'code': int(info['retcode']),
            'msg': info['errmsg'],
            'success': not bool(int(info['retcode']))
        }
        return result

    @login_required
    def recharge_net(self, amount):
        """
        充值网费（一卡通余额 ==> 城市热点）

        :param amount: 充值金额, 单位为元
        :type amount: float or int
        :return: 充值结果

        >>> card.recharge_net(11)
        {'success': True, 'code' : 0, 'Msg' : '充值成功'}
        """
        data = {
            'paytype': 1,
            'aid': AIDS['net'],
            'account': self._get_inner_account(),
            'tran': int(amount * 100),
            'netacc': '{"netacc": "", "bal": null, "pkgid": null, "lostflag": null, "freezeflag": null, '
                      '"expflag": null,"statustime": null, "duration": null, "starttime": null, "pkgtab": []}',
            'pkgflag': 'none',
            'pkg': '{}',
            'acctype': '###',
            'qpwd': '',
            'json': True
        }
        result = self.get_json(url=self.URLs.NET_PAY, method='post', data=data)
        result = json.loads(result['Msg'])
        return {
            'success': not int(result['pay_net_gdc']['retcode']),
            'code': int(result['pay_net_gdc']['retcode']),
            'msg': result['pay_net_gdc']['errmsg']
        }

    def _get_build_ids(self, aid):
        """获取建筑的id"""
        data = {
            'jsondata': '{"query_elec_building": {"aid": "%s", "account": "%s", "area": {"area": "", "areaname": ""}}}'
                        % (aid, self._get_inner_account()),
            'funname': 'synjones.onecard.query.elec.building',
            'json': True,
        }
        result = self.get_json(data=data, url=self.URLs.COMMON, method='post')
        result = json.loads(result['Msg'])
        # - -!
        building_id = {}
        for build in result['query_elec_building']['buildingtab']:
            building_id[build['building']] = build['buildingid']
        return building_id

    @login_required
    def recharge_xianlin_elec(self, amount, building_name, big_room_id, small_room_id=0):
        """
        充值仙林校区的寝室电费

        :param float amount: 充值金额
        :param str building_name: 楼栋名称，例如 "兰苑11栋"
        :param big_room_id: 大寝寝室号
        :type big_room_id: int or str
        :param small_room_id: 小寝寝室号， 不传或传0则表示充值为大寝电费， 1、2、3则充值对应小寝空调电费
        :type small_room_id: int or str
        :return: 充值结果
        :raise: :class:`njupt.exceptions.NjuptException` 楼栋名称不正确

        >>> card.recharge_xianlin_elec(
        >>>    amount=11,
        >>>    building_name='兰苑11栋',
        >>>    big_room_id=403,
        >>>    small_room_id=1
        >>> )
        {'code': 0, 'msg': '缴费成功！', 'success': True}
        """
        try:
            buiding_id = self._get_build_ids(AIDS['elec_xianlin'])[building_name]
            return self._recharge_electricity(amount=amount, elec_aid=AIDS['elec_xianlin'], building_id=buiding_id,
                                              building=building_name, room_id='{}{}'.format(big_room_id, small_room_id))
        except KeyError:
            raise NjuptException("不存在的楼栋")

    @login_required
    def recharge_sanpailou_elec(self, amount, building_name, room_id):
        """
        充值三牌楼校区的寝室电费，参数参考仙林校区（未测试
        """
        try:
            buiding_id = self._get_build_ids(AIDS['elec_sanpailou'])[building_name]
            return self._recharge_electricity(amount=amount, elec_aid=AIDS['elec_sanpailou'], building_id=buiding_id,
                                              building=building_name, room_id=room_id)
        except KeyError:
            raise NjuptException("不存在的楼栋")

    def _recharge_electricity(self, amount, elec_aid, building_id, building, room_id):
        data = {
            "acctype": "###",
            "paytype": 1,
            "aid": elec_aid,
            "account": self._get_inner_account(),
            "tran": int(amount * 100),
            "roomid": room_id,
            "room": "",
            "floorid": "",
            "floor": "",
            "buildingid": building_id,
            "building": building,
            "areaid": "",
            "areaname": "",
            "json": True
        }

        r = json.loads(self.get_json(self.URLs.ELEC_PAY, 'post', data=data)['Msg'])
        return {
            'success': not bool(int(r['pay_elec_gdc']['retcode'])),
            'code': int(r['pay_elec_gdc']['retcode']),
            'msg': r['pay_elec_gdc']['errmsg'],
        }

    @login_required
    def get_bill(self, start_date=(datetime.datetime.now() - datetime.timedelta(days=30)).strftime("%Y-%m-%d"),
                 end_date=datetime.datetime.now().strftime("%Y-%m-%d"), rows=100, page=1):
        """
        查询一卡通消费记录，默认为最近一个月的消费记录

        :param start_date: 如 "2017-12-27"
        :param end_date:  如 "2018-01-26"
        :param rows: 一次查询返回的最大记录数量,默认为100条记录
        :param page: 如果结果数量有多页，决定返回第几页。
        :return: 查询到的信息，格式见例子
        :rtype: dict

        >>> card.get_bill()
        {
            'recodes':
                [
                    {
                        'balance': 39.71,
                        'change': -5,
                        'comment': '未知系统,交电费',
                        'department': '仙林售电处',
                        'time': '2018-01-26 20:55:40',
                        'type': '代扣代缴',
                        'week': '星期五'
                    },
                    ...
                ],
            'total': 52
            'total_pages': 2
            'page': 1
        }
        """
        data = {
            "sdate": start_date,
            "edate": end_date,
            "account": self._get_inner_account(),
            "page": page,
            "rows": rows
        }
        temp = self.get_json(url=self.URLs.BILL, method="post", data=data)
        result = {'total': temp['total'], 'recodes': [], 'total_pages': temp['total'] // rows + 1, 'page': page}
        if temp['rows']:
            for row in temp['rows']:
                result['recodes'].append({
                    'balance': row['CARDBAL'],
                    'department': row['MERCNAME'].strip(),
                    'comment': row['JDESC'].strip(),
                    'change': row['TRANAMT'],
                    'week': row['XQ'],
                    'time': row['OCCTIME'],
                    'type': row['TRANNAME'].strip(),
                })
        return result



from enum import Enum

from njupt.base import API
from njupt.exceptions import AuthenticationException
from njupt.utils import login_required
from njupt.utils.captchas.library.libray_captcha import LibraryCaptcha


class LoginType(Enum):
    """
    图书馆的登录方式

    - STUDENT_ID: 学号
    - CARD_ID: 一卡通号
    - EMAIL: 邮箱
    """
    STUDENT_ID = 'cert_no'
    CARD_ID = 'bar_no'
    EMAIL = 'email'


class Library(API):
    """图书馆

    :param username: 用户名
    :param password: 密码
    :param login_type:  :class:`LoginType`, 默认为一卡通账号
    """

    class URLs:
        HOST = "http://202.119.228.6:8080"
        # 验证码
        CAPTCHA = HOST + '/reader/captcha.php?0.2226025362345463'
        # 登录
        LOGIN = HOST + '/reader/redr_verify.php'
        # 信息
        INFO = HOST + '/reader/redr_info_rule.php'

    def __init__(self, username, password, login_type=LoginType.CARD_ID):
        super().__init__()
        self.username = username
        self.password = password
        if self.username and self.password:
            self.login(username, password, login_type)

    def login(self, username, password, login_type):
        captcha_code = str(LibraryCaptcha(self.get_image(self.URLs.CAPTCHA)))
        data = {
            'number': username,
            'passwd': password,
            'captcha': captcha_code,
            'select': login_type.value,
            'returnUrl': ''
        }
        res = self.post(url=self.URLs.LOGIN, data=data)
        res.encoding = 'utf-8'
        if '注销' in res.text:
            self.verify = True
        else:
            raise AuthenticationException('incorrect username or password')

    @login_required
    def get_info(self):
        """基本信息

        :return: dict of user info
        :rtype: dict

        >>> library.get_info()
        {
            '姓名': '张三',
            '证件号': 'B12345678',
            '条码号': '11020151234556,
            '最大可借图书': '23',
            ...
        }
        """
        soup = self.get_soup(self.URLs.INFO)
        table = soup.select_one('#mylib_info > table')
        info = {}
        for tr in table.select('tr'):
            for td in tr.select('td'):
                text = td.text.strip()
                if not text:
                    continue
                key, value = text.split('：')
                info[key.strip()] = value.strip()
        return info


if __name__ == "__main__":
    lib = Library('110201500467700', '12')
    print(lib.get_info())



import json
import os
from io import BytesIO

import requests
from PIL import Image

from njupt.utils.captchas import distance_hanmming, build_vector, rotate_img

BLACK = 0
WHITE = 255

current_dir = os.path.dirname(os.path.abspath(__file__))


class ZhengfangCaptcha:
    """
    knn 识别正方验证码
    """

    def __init__(self, image):
        self.image = image
        self.image_pre_process()

    def image_pre_process(self):
        # 去除杂色点
        for x in range(self.image.width):
            for y in range(self.image.height):
                pix = self.image.getpixel((x, y))
                if pix == 43:
                    self.image.putpixel((x, y), WHITE)
                else:
                    self.image.putpixel((x, y), BLACK)

        # 去除单像素噪点并进行二值化(八值法)
        for x in range(self.image.width):
            for y in range(self.image.height):
                count = 0
                if x != 0 and y != 0 and x != self.image.width - 1 and y != self.image.height - 1:
                    for i in range(-1, 2):
                        for j in range(-1, 2):
                            tx = x + i
                            ty = y + j
                            if self.image.getpixel((tx, ty)) == BLACK:
                                count += 1
                if self.image.getpixel((x, y)) == WHITE and count == 8:
                    self.image.putpixel((x, y), BLACK)
        self.image = self.image.convert('1')

    def handle_split_image(self):
        # 切割验证码，返回包含四个字符图像的列表
        y_min, y_max = 0, 22
        split_lines = [5, 17, 29, 41, 53]
        ims = [rotate_img(self.image.crop([u, y_min, v, y_max])) for u, v in zip(split_lines[:-1], split_lines[1:])]
        return ims

    def crack(self):
        result = []
        # 加载数据
        with open(os.path.join(current_dir, 'image_data.json'), 'rb+') as f:
            image_data = json.load(f)
        for letter in self.handle_split_image():
            letter_vector = build_vector(letter)
            guess = []
            for image in image_data:
                for x, y in image.items():
                    guess.append((distance_hanmming(y, letter_vector), x))
            guess.sort()
            neighbors = guess[:15]  # 距离最近的十五个向量
            class_votes = {}  # 投票
            for neighbor in neighbors:
                class_votes.setdefault(neighbor[-1], 0)
                class_votes[neighbor[-1]] += 1
            sorted_votes = sorted(class_votes.items(), key=lambda x: x[1], reverse=True)
            result.append(sorted_votes[0][0])
        return ''.join(result)

    def __str__(self):
        return self.crack()


if __name__ == "__main__":
    from njupt import Zhengfang

    r = requests.get(Zhengfang.URLs.CAPTCHA)
    im = Image.open(BytesIO(r.content))
    im.show()
    print(ZhengfangCaptcha(im))



import os
import json

from PIL import Image

from njupt.utils.captchas import build_vector

if __name__ == "__main__":
    # 字符集
    letters = ['2', '3', '4', '5', '6', '7', '8', '9']
    # 将图像数据转为向量数据并保存
    image_data = []
    for letter in letters:
        for img in os.listdir('captcha_chars/{}/'.format(letter)):
            if img != "Thumbs.db" and img != ".DS_Store":
                temp = build_vector(Image.open("captcha_chars/{}/{}".format(letter, img)))
                image_data.append({letter: temp})

    with open('image_data.json', 'w') as f:
        json.dump(image_data, f)



# encoding: utf-8
__author__ = 'gaoliang'


class NjuptException(Exception):
    """Base class for exceptions in this module."""

    def __init__(self, *args, **kwargs):
        super(NjuptException, self).__init__(*args, **kwargs)


class AuthenticationException(NjuptException):
    """An Authentication Exception occurred."""


class UnauthorizedError(NjuptException):
    """Not authorized Error"""


class TemporaryBannedException(NjuptException):
    """Temporary Banned by Server"""



import math
import operator
from functools import reduce

BLACK = 0
WHITE = 255


def magnitude(concordance):
    total = reduce(operator.add, map(lambda x: x ** 2, concordance))
    return math.sqrt(total)


# 计算矢量之间的余弦距离
def distance_cos(vector1, vector2):
    sum_value = 0
    for value1, value2 in zip(vector1, vector2):
        sum_value += value1 * value2
    return 1 - sum_value / (magnitude(vector1) * magnitude(vector2))


def distance_hanmming(vector1, vector2):
    """
    计算两向量的汉明距，（向量只包含0，1时）
    """
    count = 0
    for value1, value2 in zip(vector1, vector2):
        if value1 != value2:
            count += 1
    return count


def build_vector(image, binary=True):
    """
    图像转一维特征向量
    :param image: pillow Image object with mode 1 or mode L
    :param binary: 黑白图是否生成为0，1向量
    :return: list of int
    """
    vector = []
    for pixel in image.getdata():
        if binary:
            vector.append(1 if pixel == 255 else 0)
        else:
            vector.append(pixel)
    return vector


def rotate_img(image):
    """
    根据图像在x轴方向投影大小确定字符的摆放方向
    :param image: PIL.Image object
    :return: rotated Image object
    """
    min_count = 1000
    final_angle = 0
    for angle in range(-45, 45):
        x_count = 0
        ti = image.rotate(angle, expand=True)
        for x in range(ti.width):
            for y in range(ti.height):
                if ti.getpixel((x, y)) == WHITE:
                    x_count += 1
                    break
        if x_count < min_count:
            min_count = x_count
            final_angle = angle
    image = image.rotate(final_angle, expand=False)
    return image



import json
import os
from io import BytesIO

import requests
from PIL import Image

from njupt.utils.captchas import distance_hanmming, build_vector

BLACK = 0
WHITE = 255

current_dir = os.path.dirname(os.path.abspath(__file__))


class LibraryCaptcha:

    def __init__(self, image):
        self.image = image.convert('1')

    def handle_split_image(self):
        # 切割验证码，返回包含四个字符图像的列表
        y_min, y_max = 16, 26
        split_lines = [4, 16, 28, 40, 52]
        ims = [self.image.crop([u, y_min, v, y_max]) for u, v in zip(split_lines[:-1], split_lines[1:])]
        return ims

    def crack(self):
        result = []
        # 加载数据
        with open(os.path.join(current_dir, 'image_data.json'), 'r') as f:
            image_data = json.load(f)
        for letter in self.handle_split_image():
            letter_vector = build_vector(letter)
            guess = []
            for image in image_data:
                for x, y in image.items():
                    guess.append((distance_hanmming(y, letter_vector), x))
            guess.sort()
            result.append(guess[0][1])
        return ''.join(result)

    def __str__(self):
        return self.crack()


if __name__ == "__main__":
    from njupt import Library

    r = requests.get(Library.URLs.CAPTCHA)
    im = Image.open(BytesIO(r.content))
    im.show()
    captcha = LibraryCaptcha(image=im)
    print(captcha.crack())



# encoding: utf-8
import re
from datetime import datetime

from bs4 import BeautifulSoup

from njupt.base import API
from njupt.exceptions import AuthenticationException, TemporaryBannedException
from njupt.utils import ZhengfangCaptcha, table_to_list, table_to_dict, login_required

week_re = re.compile(r'第(\d+)-(\d+)周')
courser_index_re = re.compile(r'第(\d+(?:,\d+)*)节')
chinese_rome = {
    '一': 1,
    '二': 2,
    '三': 3,
    '四': 4,
    '五': 5,
    '六': 6,
    '七': 7
}


class Zhengfang(API):
    """南邮正方教务

    :param str account: 南邮学号
    :param str password: 密码
    :raise: :class:`njupt.exceptions.AuthenticationException`

    >>> zf = Zhengfang(account='B11111111', password='xxx')
    """

    class URLs:
        HOST = "http://jwxt.njupt.edu.cn"

        SCORE = HOST + '/xscj_gc.aspx?xh={account}&xm=%B8%DF%C1%C1&gnmkdm=N121605'
        SCHEDULE = HOST + '/xskbcx.aspx?xh={account}&xm=%B8%DF%C1%C1&gnmkdm=N121603'
        GRADE = HOST + '/xsdjkscx.aspx?xh={account}&xm=%B8%DF%C1%C1&gnmkdm=N121606'
        SELECTED_COURSES = HOST + '/xsxkqk.aspx?xh={account}&xm=%B8%DF%C1%C1&gnmkdm=N121615'
        OPTIONAL_COURSES = HOST + '/xsxk.aspx?xh={account}&xm=%B8%DF%C1%C1&gnmkdm=N121101'
        COURSE_INFO = HOST + '/kcxx.aspx?xh={account}&kcdm={course_code}&xkkh=%26nbsp%3bk'
        # 教务系统验证码
        CAPTCHA = HOST + '/CheckCode.aspx'
        # 教务系统登录
        LOGIN = HOST + '/default2.aspx'

    def __init__(self, account=None, password=None):
        super(Zhengfang, self).__init__()
        if account and password:
            self.account = account
            self.login(account, password)

    @login_required
    def list_exam_grades(self):
        """ 获取等级考试成绩信息

        :rtype: list of dict
        :return: 列表，内容为每次等级考试的信息

        >>> zf.list_exam_grades()
        [
            {
                '学年': '2015-2016',
                '学期': 1,
                '等级考试名称': '全国大学英语四级考试',
                '准考证号': '320082152113313',
                '考试日期': datetime.datetime(2015, 12, 19, 0, 0),
                '成绩': '710'
                '写作成绩': '',
                '综合成绩': ''
            },
            ...
        ]
        """
        soup = self.get_soup(method='get', url=self.URLs.GRADE.format(account=self.account))
        table = soup.select_one('#DataGrid1')
        result = table_to_list(table, index_cast_dict={
            1: int,
            4: lambda x: datetime.strptime(x, '%Y%m%d')
        })
        return result

    @login_required
    def get_gpa(self):
        """获取GPA

        :rtype: int

        >>> zf.get_gpa()
        5.0
        """
        view_state = self._get_viewstate(url=self.URLs.SCORE.format(account=self.account))
        data = {
            'ddlXN': '',
            'ddlXQ': '',
            '__VIEWSTATE': view_state,
            'Button2': '%D4%DA%D0%A3%D1%A7%CF%B0%B3%C9%BC%A8%B2%E9%D1%AF'
        }
        soup = self.get_soup(method='post', url=self.URLs.SCORE.format(account=self.account), data=data)
        return float(soup.select_one('#pjxfjd').text[7:])

    @login_required
    def get_schedule(self, week, year=None, semester=None):
        """
        获取指定学期指定周的课表（不传入年份和学期则默认当前学期), 不推荐使用

        :param year: 学年 格式为 "2017-2018"
        :param semester: 学期 数字1或2
        :param week: 周次 数字 1-20
        :return: 二维列表schedule，schedule[i][j]代表周i第j节课的课程。 为了方便，i或j为零0的单元均不使用。
                列表的元素为None，代表没有课程，或描述课程信息的dict

        >>> zf.get_schedule(week=1, year='2017-2018', semester=1)
        [
            {
                'classroom': '教4－202',
                'name': '技术经济学',
                'teacher': '储成祥'
            }
        ]
        """
        schedule = [[None for col in range(13)] for row in range(8)]
        if year and semester:
            pass
        else:
            r = self.get(url=self.URLs.SCHEDULE.format(account=self.account))
            soup = BeautifulSoup(r.text.replace('<br>', '\n'), 'lxml')
            trs = soup.select("#Table1 > tr")
            for index, tr in enumerate(trs):
                tds = tr.select('td')
                for td in tds:
                    if len(td.text) > 4:  # 筛选出包含课程信息的单元格
                        for courser in td.text.split('\n\n'):
                            info = courser.split()
                            start_week, end_week = map(int, week_re.search(info[1]).groups())
                            courser_index = map(int, courser_index_re.search(info[1]).groups()[0].split(','))
                            is_odd_week_only = "单周" in info[1]
                            is_even_week_only = "双周" in info[1]
                            week_day = chinese_rome[info[1][1]]
                            courser = {
                                'name': info[0],
                                'teacher': info[2],
                                'classroom': info[3],
                            }
                            if start_week <= week <= end_week:
                                if (week % 2 == 0 and is_odd_week_only) or (week % 2 == 1 and is_even_week_only):
                                    pass
                                else:
                                    for courser_index in courser_index:
                                        schedule[week_day][courser_index] = courser
        return schedule

    @login_required
    def get_courses(self):
        """获取这学期的选课情况

        :return:  格式参看例子，期中interval为1则为单周，2则为双周。 这些信息足够生成课表了
        :rtype: list of dict

        >>> zf.get_courses()
        [
            {
                'class_start': 8,
                'class_end': 9,
                'day': 1,
                'name': '市场营销',
                'room': '教4－101',
                'teacher': '王波(男)',
                'week': '第1-15周|单周',
                'interval': 2,
                'week_end': 15,
                'week_start': 1
            },
            ...
        ]
        """
        r = self.get(url=self.URLs.SELECTED_COURSES.format(account=self.account))
        # soup = self.get_soup(method='get', url=self.URLs.SELECTED_COURSES.format(account=self.account))
        # 编码参考：
        # http://bbs.chinaunix.net/thread-3610023-1-1.html
        soup = BeautifulSoup(r.content, fromEncoding="gb18030")
        trs = soup.select('#DBGrid > tr')[1:]
        courses = []
        for tr in trs:
            tds = [node.text.strip() for node in tr.select('td')]
            name = tds[2]
            teacher = tds[5]
            all_time = tds[8].split(';')
            all_room = tds[9].split(';')
            for time, room in zip(all_time, all_room):
                if time and room:
                    week_start, week_end = map(int, week_re.search(time).groups())
                    courser_index = list(map(int, courser_index_re.search(time).groups()[0].split(',')))
                    week = re.search('{(.*)}', time).groups()[0]
                    if '双周' in week and week_start % 2 == 1:
                        week_start += 1
                    if '单周' in week and week_start % 2 == 0:
                        week_start += 1
                    courses.append(
                        {
                            'day': chinese_rome[time[1]],
                            'name': name,
                            'week': week,
                            'week_start': week_start,
                            'week_end': week_end,
                            'class_start': courser_index[0],
                            'class_end': courser_index[-1],
                            'teacher': teacher,
                            'room': room,
                            'interval': 2 if '单周' in week or '双周' in week else 1,
                        }
                    )
        return courses

    @login_required
    def list_exam_scores(self):
        """获取参加过的考试的成绩列表

        :rtype: list[dict]
        :return: 返回一个包含考试成绩信息字典的列表, 注意是所有参加过的考试

        >>> zf.list_exam_scores()
        [
            {
                '备注': '',
                '学分': 3.0,
                '学年': '2016-2017',
                '学期': 1,
                '学院名称': '电子科学与工程学院',
                '成绩': 90.0,
                '绩点': 4.0,
                '补考成绩': '',
                '课程代码': 'B0400111S',
                '课程名称': '模拟电子线路C',
                '课程归属': '',
                '课程性质': '必修',
                '课程英文名称': '',
                '辅修标记': '0',
                '重修成绩': '',
                '重修标记': '0'
            },
            ...
        ]

        """
        viewstate = self._get_viewstate(url=self.URLs.SCORE.format(account=self.account))
        data = {
            'ddlXN': '',
            'ddlXQ': '',
            '__VIEWSTATE': viewstate,
            'Button2': '%D4%DA%D0%A3%D1%A7%CF%B0%B3%C9%BC%A8%B2%E9%D1%AF'
        }
        soup = self.get_soup(method='post', url=self.URLs.SCORE.format(account=self.account), data=data)
        table = soup.select_one('#Datagrid1')
        return table_to_list(table, index_cast_dict={
            1: int,
            6: float,
            7: lambda x: float(x) if x else None,
            8: lambda x: float(x) if x.isdigit() else x
        })

    @login_required
    def get_gpa_under_pku(self):
        """获取按照北大GPA算法计算的绩点

        :return: 北大算法绩点，注意此方法不计算任选课的成绩

        >>>zf.get_gpa_under_pku()
        """
        scores = self.list_exam_scores()
        effective_courses = [score for score in scores if score['课程性质'] != '任选']
        total_credits = 0
        academic_credits = 0
        for score in effective_courses:
            if score['成绩'] == '优秀':
                score['成绩'] = 90
            elif score['成绩'] == '良好':
                score['成绩'] = 80
            elif score['成绩'] == '中等':
                score['成绩'] = 70
            elif score['成绩'] == '合格':
                score['成绩'] = 60
            elif score['成绩'] == '不合格':
                score['成绩'] = 59
        for score in effective_courses:
            if score['成绩'] > 60:
                score['绩点'] = float('%.2f' % (4 - 3 * (100 - score['成绩']) ** 2 / 1600))
            else:
                if score['补考成绩'] == '及格':
                    score['绩点'] = 1.0
                else:
                    score['绩点'] = 0.0
            academic_credits += score['学分'] * score['绩点']
            total_credits += score['学分']
        return float('%.2f' % (academic_credits / total_credits))

    @login_required
    def get_gpa_under_zju(self):
        """获取按照浙大GPA算法计算的绩点

        :return: 浙大算法绩点，注意此方法不计算任选课的成绩

        >>> zf.get_gpa_under_zju()

        """
        scores = self.list_exam_scores()
        effective_courses = [score for score in scores if score['课程性质'] != '任选']
        total_credits = 0
        academic_credits = 0
        for score in effective_courses:
            if score['成绩'] == '优秀':
                score['成绩'] = 90
            elif score['成绩'] == '良好':
                score['成绩'] = 80
            elif score['成绩'] == '中等':
                score['成绩'] = 70
            elif score['成绩'] == '合格':
                score['成绩'] = 60
            elif score['成绩'] == '不合格':
                score['成绩'] = 59
        for score in effective_courses:
            if score['成绩'] >= 85:
                score['绩点'] = 4.0
            elif 60 <= score['成绩'] <= 84:
                score['绩点'] = (score['成绩'] - 60) * 0.1 + 1.5
            else:
                if score['补考成绩'] == '及格':
                    score['绩点'] = 1.0
                else:
                    score['绩点'] = 0.0
            academic_credits += score['学分'] * score['绩点']
            total_credits += score['学分']
        return float('%.2f' % (academic_credits / total_credits))

    @login_required
    def list_optional_courses(self):
        """获取可选课程列表，对应于教务系统 -> 网上选课 -> 学生选课

        :rtype: list of dict
        :return: 可选课程信息列表
        :raise: :class:`njupt.exceptions.TemporaryBannedException`
        """
        soup = self.get_soup(self.URLs.OPTIONAL_COURSES.format(account=self.account))
        if len(str(soup)) < 100:
            raise TemporaryBannedException('选课三秒防刷')
        table = soup.select_one('#kcmcgrid')
        result = table_to_list(table, remove_index_list=[8], index_cast_dict={
            4: int,
            9: int
        })
        return result[:-1]  # 最后多了一个空行

    @login_required
    def get_course_info(self, courser_code):
        """
        获取课程的简介信息

        :param courser_code: 课程代码
        :rtype: dict

        """
        soup = self.get_soup(self.URLs.COURSE_INFO.format(account=self.account, course_code=courser_code))
        table = soup.select_one('#Table1')
        return table_to_dict(table)

    def login(self, account, password):
        # FIXME: 登录不可靠
        captcha_code = ZhengfangCaptcha(self.get_image(self.URLs.CAPTCHA)).crack()
        data = {
            "__VIEWSTATE": self._get_viewstate(self.URLs.LOGIN),
            'txtUserName': account,
            'TextBox2': password,
            'RadioButtonList1': "%D1%A7%C9%FA",
            "txtSecretCode": captcha_code,
            "Button1": "",
            "hidPdrs": "",
            "hidsc": ""
        }
        result = self._login_execute(url=self.URLs.LOGIN, data=data)
        if result['code'] == 2:
            # 如果验证码错误，尝试递归重复登录
            return self.login(account, password)
        result['success'] = not result['code']
        if result['success']:
            self.verify = True
        else:
            raise AuthenticationException(result['msg'])
        return result

    def _login_execute(self, url=None, data=None):
        r = self.post(url=url, data=data)
        if r.ok:
            if "请到信息维护中完善个人联系方式" in r.text:
                self.account = data['txtUserName']
                # self.cookies.save(ignore_discard=True)  # 保存登录信息cookies
                self.verify = True  # 登陆成功, 修改状态  (后期还可能继续修改)
                # self.cookies.load(filename=settings.COOKIES_FILE, ignore_discard=True)
                return {'code': 0, 'msg': '登录成功'}
            elif "密码错误！！" in r.text:
                return {'code': 1, 'msg': '密码错误！！'}
            elif "验证码不正确！！" in r.text:
                return {'code': 2, 'msg': '验证码不正确！！！'}
            else:
                return {'code': 3, 'msg': '未知错误'}
        else:
            return {'code': 1, "msg": "登录失败"}



"""
奥兰系统验证码识别，利用向量空间余弦相似度实现
"""
import os
import pickle
from io import BytesIO

from PIL import Image

# 对图像进行纵向分割的分割点
from njupt.utils.captchas import distance_cos, build_vector

letters = [(3, 11), (23, 31), (43, 51), (63, 71)]
current_dir = os.path.dirname(os.path.abspath(__file__))

BLACK = 0
WHITE = 255


class AolanCaptcha:
    # 分割点
    letters = [(3, 11), (23, 31), (43, 51), (63, 71)]

    def __init__(self, im):
        # 路径或BytesIO或文件时
        if isinstance(im, str) or isinstance(im, BytesIO) or type(im) == "file":
            self.im = Image.open(im)
        elif isinstance(im, Image.Image):
            self.im = im
        else:
            raise TypeError("错误的类型")
        self.binarize()

    def binarize(self):
        # 灰度处理
        self.im = self.im.convert('L')

        # 二值处理
        threshold = 220
        table = []
        for i in range(256):
            if i < threshold:
                table.append(BLACK)
            else:
                table.append(WHITE)
        self.im = self.im.convert('1')

    def crack(self):
        # 装载训练数据集
        with open(os.path.join(current_dir, 'imageset.dat'), 'rb+') as f:
            imageset = pickle.load(f)
        # 对验证码图片进行切割
        result = []
        for letter in letters:
            final = self.im.crop((letter[0], 0, letter[1], self.im.size[1]))
            letter_vector = build_vector(final)
            guess = []
            # 将切割得到的验证码小片段与每个训练片段进行比较
            for image in imageset:
                for x, y in image.items():
                    guess.append((distance_cos(y, letter_vector), x))
            guess.sort()
            result.append(guess[0][1])
        return "".join(result)

    def __str__(self):
        return self.crack()


import hashlib
import json
import os
import shutil
import time

from PIL import Image

from njupt.utils.captchas import build_vector
from njupt.utils.captchas.card import CardCaptcha

BLACK = 0
WHITE = 255


def spilt2chars():
    """
    分割已有的数据为字符并保存
    """
    try:
        shutil.rmtree('captcha_chars')
    except FileNotFoundError:
        pass
    os.mkdir("captcha_chars")
    values = "1234567890"
    for value in values:
        os.mkdir('captcha_chars/{}'.format(value))

    file_names = os.listdir('captchas')
    for file_name in file_names:  #
        if not os.path.isdir('captchas/{}'.format(file_name)) and file_name != '.DS_Store':
            values = file_name[:4]
            im = Image.open('captchas/{}'.format(file_name))
            captcha = CardCaptcha(im)
            for im_part, value in zip(captcha.handle_split_image(), values):
                m = hashlib.md5()
                m.update("{}{}".format(time.time(), value).encode('utf8'))
                im_part.save("captcha_chars/{}/{}.png".format(value, m.hexdigest()))


if __name__ == "__main__":
    # spilt2chars()
    iconset = list('0123456789')
    # # 将图像数据转为向量数据并保存
    image_data = []
    for letter in iconset:
        try:
            for img in os.listdir('captcha_chars/{}/'.format(letter)):
                if img != "Thumbs.db" and img != ".DS_Store":
                    temp = build_vector(Image.open("captcha_chars/{}/{}".format(letter, img)))
                    image_data.append({letter: temp})
        except FileNotFoundError:
            print(letter)

    with open('image_data.json', 'w') as f:
        json.dump(image_data, f)



import json
import os

from PIL import Image

from njupt.utils.captchas import distance_hanmming, build_vector

BLACK = 0
WHITE = 255

current_dir = os.path.dirname(os.path.abspath(__file__))


class CardCaptcha:
    """
    一卡通系统的验证码工具类
    """

    def __init__(self, im):
        """
        :param im: PIL.image对象
        """
        self.im = im.convert('L')
        self.image_pre_process()

    def image_pre_process(self):

        threshold = 200
        # 去除杂色点
        for x in range(self.im.width):
            for y in range(self.im.height):
                pix = self.im.getpixel((x, y))
                if pix > threshold:
                    self.im.putpixel((x, y), BLACK)
                else:
                    self.im.putpixel((x, y), WHITE)

    def handle_split_image(self):
        # 切割验证码，返回包含五个字符图像的列表
        y_min, y_max = 0, 82
        split_lines = [7, 49, 91, 133, 175, 217]
        ims = [self.im.crop([u, y_min, v, y_max]) for u, v in zip(split_lines[:-1], split_lines[1:])]
        return ims

    def crack(self):
        result = []
        # 装载训练数据集
        with open(os.path.join(current_dir, 'image_data.json'), 'rb+') as f:
            image_data = json.load(f)
        for letter in self.handle_split_image():
            letter_vector = build_vector(letter)
            guess = []
            for image in image_data:
                for x, y in image.items():
                    if len(y) != 0:
                        guess.append((distance_hanmming(y, letter_vector), x))
            guess.sort()
            result.append(guess[0][1])
        return ''.join(result)

    def __str__(self):
        return self.crack()


if __name__ == "__main__":
    im = Image.open("captchas/12272.gif")
    im.show()
    print(CardCaptcha(im).crack())



from functools import wraps

from njupt.utils.captchas.zhengfang.zhengfang_captcha import ZhengfangCaptcha  # noqa
from njupt.exceptions import UnauthorizedError


def table_to_list(table, remove_index_list=None, index_cast_dict=None):
    """将html的table转换python中的list

    :type table
    :param table: bs4 选中的table
    :type remove_index_list: list[int] or None
    :param remove_index_list: 需要剔除的列的下标
    :type index_cast_dict: dict[int, function] or None
    :param index_cast_dict: 类型转换函数集， key为列的下标，value为对该值的cast函数
    :rtype list[dict]
    :return: table转换而成的列表， 每个元素为包含所需信息的字典
    """
    index_cast_dict = index_cast_dict or {}
    remove_index_list = remove_index_list or []
    trs = list(table.select('tr'))
    keys = [key.text.strip() for col_index, key in enumerate(trs[0].select('td')) if col_index not in remove_index_list]
    result = []
    for tr in trs[1:]:
        values = []
        for col_index, td in enumerate(tr.select('td')):
            if col_index in remove_index_list:
                continue
            value = td.text.strip()
            if col_index in index_cast_dict:
                value = index_cast_dict[col_index](value)
            values.append(value)
        recode = dict(zip(keys, values))
        result.append(recode)
    return result


def table_to_dict(table, remove_index_list=None, index_cast_dict=None):
    """将html的table转换python中的dict

    :type table
    :param table: bs4 选中的table
    :type remove_index_list: list[int] or None
    :param remove_index_list: 需要剔除的列的下标
    :type index_cast_dict: dict[int, function] or None
    :param index_cast_dict: 类型转换函数集， key为列的下标，value为对该值的cast函数
    :rtype list[dict]
    :return: table转换而成的列表， 每个元素为包含所需信息的字典
    """
    index_cast_dict = index_cast_dict or {}
    remove_index_list = remove_index_list or []
    trs = table.select('tr')
    result = {}
    for row_index, tr in enumerate(trs):
        if row_index in remove_index_list:
            continue
        tds = tr.select('td')
        if len(tds) != 2:
            continue
        key = tds[0].text.strip()
        value = tds[1].text.strip()
        if row_index in index_cast_dict:
            value = index_cast_dict[row_index](value)
        result[key] = value
    return result


def login_required(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.verify:
            return func(self, *args, **kwargs)
        else:
            raise UnauthorizedError('login required!')
    return wrapper
